#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <GL/glut.h>

#include "rigidXform.h"

#include "randj.h"

#define MIN(x,y)	(((x)<(y))?(x):(y))
#define MAX(x,y)	(((x)>(y))?(x):(y))

#define DEG_TO_RAD   (M_PI/180.0)

#define FLOAT_GL GLdouble
#define GLVERTEX glVertex3d



static GLsizei CurrentWindowSize[2] = {907, 907};


// Note that this currently assumes a symmetric frustum
typedef struct camera_struct
	{
	double location[3];
	double orientation[4];
	GLdouble fov;
	GLdouble near, far;
	} Camera;

typedef struct view_struct
	{
	int idNum;
	Camera camera;
	GLsizei windowSize[2];
	} View;


/////////////////////////////////////////////////////////////////////

static void
printCamera (FILE *fp, Camera *camera)
	{
	fprintf (fp, "Camera:\n");
	fprintf (fp, "    location %f %f %f\n", 
			camera->location[0],
			camera->location[1],
			camera->location[2]);
	fprintf (fp, "    orientation %f %f %f %f\n",
			camera->orientation[0],
			camera->orientation[1],
			camera->orientation[2],
			camera->orientation[3]);
	fprintf (fp, "    fov %f\n", camera->fov);
	fprintf (fp, "    near %f    far %f\n", camera->near, camera->far);

	}

static void
printView (FILE *fp, View *view)
	{
	fprintf (fp, "VIEW    idNum = %d\n", view->idNum);
	printCamera (fp, &(view->camera));
	fprintf (fp, "windowSize = %d %d\n", 
			view->windowSize[0], view->windowSize[1]);
	fprintf (fp, "\n");
	}


////////////////////////////////////////////////////////////////

//
// In this section we are going to initialize a bunch of test
// cases for checking the point rendering errors.  This will
// involve generating a set of test points, and test viewing
// parameters.
//

static int NumTestViews = 8;
static View *TestViews = NULL;
static int NumTestPoints = 10;
static double *TestPoints = NULL;

static void
initRenderingTestCases ()
	{
	int i, j; 
#if 1
	int iv, iwin, ifov, ior;
	int winSize[2][2] = { {128, 128}, {907, 907} };
	double fov[2] = {30.0, 150.0};
	double orientation[2][4];
	double len;
	double *pt;
#endif

	if ((TestViews = (View *) malloc(NumTestViews * sizeof (View))) == NULL)
		{
		fprintf (stderr, "Unable to allocate memory for test views.\n");
		exit (-1);
		}

        if ((TestPoints = (double *)
                          malloc (3 * NumTestPoints * sizeof (double))) == NULL)                {
                fprintf(stderr, "Unable to allocate memory for test points.\n");                exit (-1);
                }




	// generate views
	ustart (4339);  // seed for views

	for (ior = 0; ior < 2; ior++)
		{
		len = 0;
		for (j = 0; j < 4; j++)
			{
			orientation[ior][j] = uni ();
			len += orientation[ior][j] * orientation[ior][j];
			}

		len = sqrt (len);

		for (j = 0; j < 4; j++)
			{
			orientation[ior][j] /= len;
			}
		}


	iv = 0;
	for (iwin = 0; iwin < 2; iwin++)
	    {
	    for (ifov = 0; ifov < 2; ifov++)
		{
		for (ior = 0; ior < 2; ior++)
			{
			TestViews[iv].idNum = iv;
			TestViews[iv].windowSize[0] = winSize[iwin][0];
			TestViews[iv].windowSize[1] = winSize[iwin][1];
			TestViews[iv].camera.fov = fov[ifov];
			TestViews[iv].camera.orientation[0]=orientation[ior][0];
			TestViews[iv].camera.orientation[1]=orientation[ior][1];
			TestViews[iv].camera.orientation[2]=orientation[ior][2];
			TestViews[iv].camera.orientation[3]=orientation[ior][3];

			TestViews[iv].camera.location[0] = 0;
			TestViews[iv].camera.location[1] = 0;
			TestViews[iv].camera.location[2] = 0;
			TestViews[iv].camera.near = 0.001;
			TestViews[iv].camera.far = 20.0;

			iv++;
			}
		}
	    }
	

	// generate points
	ustart (5019);  // seed for points

        pt = TestPoints;
        for (i = 0; i < NumTestPoints; )
                {
                pt[0] = -1.0 + 2 * uni ();
                pt[1] = -1.0 + 2 * uni ();
                pt[2] = -1.0 + 2 * uni ();

                if ( ( pt[0]*pt[0] + pt[1]*pt[1] + pt[2]*pt[2] ) <= 1.0 )
                        {
                        // We want only points inside the unit sphere
                        pt += 3;
                        i++;
                        }
                }




	}  // end of initRenderingTestCases


static int CurrentCase = 0;


static void
ungetLastRenderingTestCase ()
	{
	if (CurrentCase > 0)
		{
		CurrentCase--;
		}
	}

static int
getNextRenderingTestCase (View *view, double pt[3])
	{
	int currentView, currentPt;

	if (CurrentCase >= (NumTestViews * NumTestPoints))
		{
		return -1;
		}



	currentView = CurrentCase / NumTestPoints;
	currentPt = CurrentCase % NumTestPoints;

	memcpy (view, TestViews+currentView, sizeof (View));

        pt[0] = TestPoints[3*currentPt + 0];
        pt[1] = TestPoints[3*currentPt + 1];
        pt[2] = TestPoints[3*currentPt + 2];

	CurrentCase++;


	if ((CurrentCase % 1000) == 0)
		{
		fprintf (stderr, "At case %d\n", CurrentCase);
		}

	return 0;
	}  // end of getNextRenderingTestCase




////////////////////////////////////////////////////////////////

static void setupViewing (View *view)

	{
	GLdouble aspectRatio;
	double cameraMat[4][4];
	

	//
	// Doing this routine every time (rather than only when the
	// viewing changes) seems to degrade performance by only
	// about 7% on a 500x500 window (on acadie).
	//

	if ( (view->windowSize[0] > CurrentWindowSize[0]) ||
	     (view->windowSize[1] > CurrentWindowSize[1])     )
		{
		fprintf (stderr, "Specified viewport size (%d, %d) "
		  "is greater than current window size (%d, %d).\n",
		  view->windowSize[0], view->windowSize[1],
		  CurrentWindowSize[0], CurrentWindowSize[1]);
		exit (-1);
		}

#if 0
	CurrentViewportSize[0] = view->windowSize[0];
	CurrentViewportSize[1] = view->windowSize[1];
#endif

   	glViewport (0, 0, view->windowSize[0], view->windowSize[1] );

	// If this divide fails because windowSize[1] == 0, then you're getting
	// what you deserve.
	aspectRatio = 
	  ((double)view->windowSize[0]-1) / ((double)view->windowSize[1]-1);
	
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();
	gluPerspective (view->camera.fov, aspectRatio, 
				view->camera.near, view->camera.far);

	glMatrixMode (GL_MODELVIEW);

	glLoadIdentity ();

	quatToMat (view->camera.orientation, cameraMat);
	gluLookAt (
			view->camera.location[0],
			view->camera.location[1],
			view->camera.location[2],
			-cameraMat[2][0],
			-cameraMat[2][1],
			-cameraMat[2][2],
			cameraMat[1][0],
			cameraMat[1][1],
			cameraMat[1][2]   );


	}  // end of setupViewing


void init (void) 
{
   glClearColor (0.0, 0.0, 0.0, 0.0);
   glShadeModel (GL_FLAT);

   initRenderingTestCases ();
}




/////////////////////////////////////////////////////////////////

static int 
getRenderedCoords (GLsizei windowSize[2], int pixelXY[2]) 
	{
	unsigned char *pixel;
	int i, j;
	int pixelFound;
	static unsigned char *Scrbuf = NULL;
	static int ScrbufSize = 0;
	int newScrbufSize;




        newScrbufSize =
          4 * windowSize[0] * windowSize[1] * sizeof(unsigned char);
        if (ScrbufSize < newScrbufSize)
                {
                if (Scrbuf != NULL)
                        {
                        free (Scrbuf);
                        }


                ScrbufSize = newScrbufSize;
                Scrbuf = (unsigned char *) malloc (ScrbufSize);

                if (Scrbuf == NULL)
                        {
                        fprintf (stderr,
                        "getRenderedCoords: Fatal error allocating memory.\n");
                        exit (-1);
                        }
                }



	glReadPixels (0, 0, windowSize[0], windowSize[1], 
				GL_RGBA, GL_UNSIGNED_BYTE, Scrbuf);


	pixelFound = 0;
	pixel = Scrbuf;
	for (j = 0; j < windowSize[1]; j++)
		{
		for (i = 0; i < windowSize[0]; i++)
			{
			if (pixel[0] != 0)
				{

				if (pixelFound)
					{
					fprintf (stderr, 
					  "getRenderedCoords: multiple pixels "
					  "found in same image.\n");
					glFinish();
					glutReportErrors ();
					exit (-1);
					}
				pixelFound = 1;
				pixelXY[0] = i;
				pixelXY[1] = j;
				}
			pixel += 4;
			}
		}




	return ! pixelFound; 
	}  // end of getRenderedCoords


static int
getCalculatedCoords ( View *view, FLOAT_GL point[3], double screenXY[2] )
	{


	// Not the most efficient implementation, but it doesn't have to be.

	double tmpPt[3];
	double camToWorldRot[4][4];
	double worldToCamRot[4][4];
	double pointInCamCoords[3];
	double projectedCamXY[2];
	double aspectRatio;
	double botEdge, vertExtent;
	double leftEdge, horizExtent;

	tmpPt[0] = point[0] - view->camera.location[0];
	tmpPt[1] = point[1] - view->camera.location[1];
	tmpPt[2] = point[2] - view->camera.location[2];

	quatToMat (view->camera.orientation, camToWorldRot);
	transposeMat (camToWorldRot, worldToCamRot);

	xformPointByMat (tmpPt, worldToCamRot, pointInCamCoords);
	

	// So the camera is at the origin looking down the -Z axis 
	// with Y up and X right.
	//
	// We now calculate the XY coordinates of the projection of the
	// pointInCamCoords on the plane z = -1;

	projectedCamXY[0] = - pointInCamCoords[0] / pointInCamCoords[2];
	projectedCamXY[1] = - pointInCamCoords[1] / pointInCamCoords[2];

	if (pointInCamCoords[2] >= 0)
		{
		// then the point is behind the camera and cannot be
		// properly projected onto the screen.
		return -1;
		}



	// These camera coordinates are aligned with the screen 
	// pixel coordinates.

	// Now we must transform the camera coords into pixel coordinates.


	// Remember: we are handling only symmetric viewing frusta.

	// The angular field of view specifies the vertical field
	// of view for the window.  So the vertical extent of the
	// the window on the screen corresponds to this Y interval
	// on the z=-1 projection plane: [-cos(fov/2), cos(fov/2)]
	// and this interval is mapped linearly to the pixel 
	// interval [0, windowSize[1]].

#define DEG_TO_RAD   (M_PI/180.0)

	botEdge = - tan (DEG_TO_RAD*view->camera.fov/2.0);
	vertExtent = 2 * tan (DEG_TO_RAD*view->camera.fov/2.0);



	aspectRatio = ((double)(view->windowSize[0]-1)) / 
				((double)(view->windowSize[1]-1));
	horizExtent = vertExtent * aspectRatio;
	leftEdge = botEdge * aspectRatio;



	// It appear that this IS the way that OpenGL maps to pixel coords
	// on : visage
	screenXY[0] = -0.5 + ((projectedCamXY[0] - leftEdge) / 
				horizExtent) * (view->windowSize[0]);
	screenXY[1] = -0.5 + ((projectedCamXY[1] - botEdge) / 
				vertExtent) * (view->windowSize[1]);



	return 0;

	}  // end of getCalculatedCoords



#define TRUE_VISIBLE            0
#define FALSE_VISIBLE           1
#define TRUE_INVISIBLE          2
#define FALSE_INVISIBLE         3


static int 
getPointRenderError (View *view, FLOAT_GL point[3], double *error, double xyError[2])
	{
	int pixelXY[2];
	double screenXY[2];
	int renderedPtVisible, calculatedPtIsValid;
	int rtn;

	setupViewing ( view );

	// Begin rendering

	// Clear the frame buffer
	glClear (GL_COLOR_BUFFER_BIT);


	glColor3f (1.0, 1.0, 1.0);
	glBegin (GL_POINTS);
	    GLVERTEX ( point[0], point[1], point[2]);
	glEnd ();


	glFinish ();


	// End rendering



	renderedPtVisible = ! getRenderedCoords (view->windowSize, pixelXY) ;


	calculatedPtIsValid = ! getCalculatedCoords ( view, point, screenXY);


	if (renderedPtVisible)
		{

		// We found a rendered pixel

		if (calculatedPtIsValid)
			{
			xyError[0] = (screenXY[0] - ((double)pixelXY[0]));
			xyError[1] = (screenXY[1] - ((double)pixelXY[1]));

			*error = sqrt 
					( 
	  				(screenXY[0] - ((double)pixelXY[0])) * 
					(screenXY[0] - ((double)pixelXY[0]))
					+
	  				(screenXY[1] - ((double)pixelXY[1])) * 
					(screenXY[1] - ((double)pixelXY[1]))
					);
			// rtn = VISIBLE_VALID;
			if ( (screenXY[0] >= -0.5) &&
				   (screenXY[0] <= view->windowSize[0]+0.5) &&
			     (screenXY[1] >= -0.5) &&
				   (screenXY[1] <= view->windowSize[1]+0.5) )
				{
				// It was correctly visible
				rtn = TRUE_VISIBLE;
				}
			else
				{
				// It was visible; but should not have been, 
				//   however we can calculate the error
				rtn = FALSE_VISIBLE;
				}
			}
		else
			{
			// The point appeared in the frame buffer but it 
			// should not have.
			*error = xyError[0] = xyError[1] = 0;
			rtn = FALSE_VISIBLE;
			}
		}
	else
		{

		// We found NO rendered pixels

		if (calculatedPtIsValid)
			{
			if ( (screenXY[0] > -0.5) &&
				   (screenXY[0] < view->windowSize[0]+0.5) &&
			     (screenXY[1] > -0.5) &&
				   (screenXY[1] < view->windowSize[1]+0.5) )
				{
				// The rendered point should be visible, but
				// it is not.
				*error = xyError[0] = xyError[1] = 0;
				rtn = FALSE_INVISIBLE;
				}
			else
				{
				// The rendered point should not be visible
				// and it wasn't.
				*error = xyError[0] = xyError[1] = 0;
				rtn = TRUE_INVISIBLE;
				}
			}
		else
			{
				// The rendered point should not be visible
				// and it wasn't.
				*error = xyError[0] = xyError[1] = 0;
				rtn = TRUE_INVISIBLE;
			}

		}

	return rtn;
	}	// end of getPointRenderError
	

static FILE *ViewFP = NULL;
static FILE *ErrorFP = NULL;

void display(void)
	{

	View view;
	double error, xyError[2];
	int rtn;
	double point[3];
	static int lastViewId = -1;
	
	// Clear out the entire window.
	glViewport (0, 0, CurrentWindowSize[0], CurrentWindowSize[1] );
	sleep (2);

	glClear (GL_COLOR_BUFFER_BIT);
	sleep (2);

	glFinish ();


	while (getNextRenderingTestCase (&view, point) == 0)
		{

		error = xyError[0] = xyError[1] = 0;

		// Render, calc errors, print errors
                rtn = getPointRenderError (&view, point, &error, xyError);
                fprintf (ErrorFP,
			" %d  %d "
                        " %lf %lf %lf "
                        " %lf %lf %lf "
                        "\n",
			view.idNum,
                        rtn,
                        point[0],
                        point[1],
                        point[2],
                        error, xyError[0], xyError[1]);

		if (view.idNum != lastViewId)
			{
			printView (ViewFP, &view);
			lastViewId = view.idNum;
			}

		}

	fclose (ErrorFP);
	fclose (ViewFP);

	exit (0);

	}  // end of display


void reshape (int w, int h)
{
   CurrentWindowSize[0] = w;
   CurrentWindowSize[1] = h;
}

void keyboard(unsigned char key, int x, int y)
{
   switch (key) {
      case 27:
         exit(0);
         break;
   }
}


static void
usage ()
	{
	fprintf (stderr, "\n");
	fprintf (stderr, 
	  "Usage: pointRenderingErrors [glut args] numPt  viewFN  errorFN\n");
	fprintf (stderr,
	 "   [glut args] See GLUT documentation.\n");
	fprintf (stderr,
	 "   numPt   number of points drawn for each view\n");
	fprintf (stderr,
	 "   viewFN   file to which a description of each view is written.\n");
	fprintf (stderr,
	 "   errorFN  file to which error data is written for each "
	 "point and each view.\n");
	fprintf (stderr, "\n");
	return;
	}

//
// command line:    progName  numPt  viewFN  errorFN
//

int main(int argc, char** argv)
{
   glutInit(&argc, argv);


   if (argc != 4)
	{
	fprintf (stderr, "Bad command line arguments.\n");
	usage ();
	exit (-1);
	}

   if (sscanf (argv[1], "%d", &NumTestPoints) != 1)
	{
	fprintf (stderr, "Bad argument: < %s >.\n", argv[1]);
	usage ();
	exit (-1);
	}

   if (NumTestPoints < 1)
	{
	fprintf (stderr, "Bad argument: < %s >.\n", argv[1]);
	usage ();
	exit (-1);
	}

   if ( (ViewFP = fopen (argv[2], "w")) == NULL)
	{
	fprintf (stderr, "Bad argument: < %s >.\n", argv[2]);
	usage ();
	exit (-1);
	}

   if ( (ErrorFP = fopen (argv[3], "w")) == NULL)
	{
	fprintf (stderr, "Bad argument: < %s >.\n", argv[3]);
	usage ();
	exit (-1);
	}



   // It seems to be faster (on acadie) to do it all in the front buffer
   glutInitDisplayMode (GLUT_SINGLE | GLUT_RGB);


   glutCreateWindow (argv[0]);

   init ();

   glutDisplayFunc(display); 
   glutReshapeFunc(reshape);
   glutKeyboardFunc(keyboard);

#if 0
   glutFullScreen ();
#else
   glutReshapeWindow (907, 907);
#endif


   glutMainLoop();
   return 0;
}
