
/*
Note when # of triangles is 25000, case 132583 seems quite anomalous.

There is no ragged edge, but the number of rendered pixels is
wildly off.  And that number changes substantially when the
window is positioned differently on the screen.
-geometry +50+50 
gives different results than
-geometry +50+51

The numbers seem to cycle with a period of 32 in x and y.  
So, for example, -geometry +50+50 gives the same results as 
-geometry +50+82 or -geometry +82+50.

*/



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <GL/glut.h>

#include "rigidXform.h"
#include "randj.h"


// static int Debug = 0;

#define DEG_TO_RAD   (M_PI/180.0)
#define MIN(x,y)  (((x)<(y))?(x):(y))
#define MAX(x,y)  (((x)>(y))?(x):(y))

#define FLOAT_GL GLdouble
#define GLVERTEX glVertex3d


//
//
// NOTE:  All polygons that we handle here are triangles or are derived
//   from triangles by clipping to half-spaces and by projection.  This
//   means that they will all be CONVEX.
//

//
// The polygons that we will have to handle here will have no more than
// maybe 9 vertices.
//
#define MAX_VERTICES (100)

typedef struct polygon
	{
	unsigned int numVerts;
	double vertex[MAX_VERTICES][3];
	int vertexFlag[MAX_VERTICES];
	} Polygon;



// Note that this currently assumes a symmetric frustum
typedef struct camera_struct
        {
        double location[3];
        double orientation[4];
        GLdouble fov;
        GLdouble near, far;
        } Camera;

typedef struct view_struct
        {
	int idNum;
        Camera camera;
        GLsizei windowSize[2];
        } View;




//
// When we talk about a 3D polygon, we mean a polygon for which the X, Y, 
// and Z coordinates are all used.  When we refer to a 2D polygon, we mean
// a polygon for which only the X and Y coordinates are used.
//
// In this program, when we say "pixel area" we are referering to the
// nominal 2D polygon covered by the pixel in pixel coordinates.  This is 
// an axis-aligned square centered at the integral pixel coordinates with 
// width and height of 1.
//


///////////////////////////////////////////////////////////////


static void
printCamera (FILE *fp, Camera *camera)
        {
        fprintf (fp, "Camera:\n");
        fprintf (fp, "    location %f %f %f\n",
                        camera->location[0],
                        camera->location[1],
                        camera->location[2]);
        fprintf (fp, "    orientation %f %f %f %f\n",
                        camera->orientation[0],
                        camera->orientation[1],
                        camera->orientation[2],
                        camera->orientation[3]);
        fprintf (fp, "    fov %f\n", camera->fov);
        fprintf (fp, "    near %f    far %f\n", camera->near, camera->far);

        }


static void
printView (FILE *fp, View *view)
        {
        fprintf (fp, "VIEW    idNum = %d\n", view->idNum);
        printCamera (fp, &(view->camera));
        fprintf (fp, "windowSize = %d %d\n",
                        view->windowSize[0], view->windowSize[1]);
        fprintf (fp, "\n");
        }


static void
printPolygon (Polygon *poly)
	{
	int i;

	printf ("POLYGON\n");
	for (i = 0; i < poly->numVerts; i++)
		{
		printf ("  %f %f %f     (%d)\n", 
				poly->vertex[i][0],
				poly->vertex[i][1],
				poly->vertex[i][2],
				poly->vertexFlag[i] );
		}
	}


#define INSIDE		0
#define OUTSIDE		1
#define BOUNDARY	2


///////////////////////////////////////////////////////////////
//
//
// Clip a 2D polygon to a vertical or horizontal half-plane
// 
//	whichComponent - Indicates which component the half-plane is dividing
//		0 means X (the dividing line is vertical), 1 means Y (the
//		dividing line is horizontal).  
//
//      bound - the boundary value for the half-plane
//
// 	direction - Indicates which side of half-plane is to be retained
//	 	 1 : include positive side
//		-1 : include negative side
//
//  The polygon data structure is modified to be the clipped polygon.
//
void
clipPolyVH (
  Polygon *poly, 
  int whichComponent, 
  double bound, 
  int direction)
	{
	int numVout;
	Polygon clippedPoly;
	unsigned int i, next_i, doIntersection;
	double d;
	int currInclusion, nextInclusion;
	int c0, c1;

	if (poly->numVerts == 0)
		{
		return;
		}

	c0 = whichComponent;
	c1 = 1 - c0;

	if (poly->numVerts == 1)
		{
		d = direction * (poly->vertex[0][c0] - bound);
		if (d <= 0)
			{
			poly->numVerts = 0;
			return;
			}
		}



	numVout = 0;



	// We step through the vertices of the polygon.  We
	// include the vertices that are in the included half
	// plane (or boundary), exclude those not.  The points
	// where the polygon edges intersect the boundary are
	// inserted in the right place.
	for (i = 0; i < poly->numVerts; i++)
		{
		
		// i and next_i are the indices of the current and next
		// vertices of the polygon.
		next_i = (i + 1) % poly->numVerts;

		// currInclusion and nextInclusion indicate whether the
		// two vertices are inside, outside, or on the boundary
		// of the half-plane.

		d = direction * (poly->vertex[i][c0] - bound);
		currInclusion = 
		   (d > 0) ? INSIDE : ((d == 0) ? BOUNDARY : OUTSIDE);

		d = direction * (poly->vertex[next_i][c0] - bound);
		nextInclusion = 
		   (d > 0) ? INSIDE : ((d == 0) ? BOUNDARY : OUTSIDE);

	

		if (currInclusion != OUTSIDE)
			{
			// If the current vertex is inside or on the boundary,
			// we include it in the clipped polygon.
			clippedPoly.vertex[numVout][c0] = poly->vertex[i][c0];
			clippedPoly.vertex[numVout][c1] = poly->vertex[i][c1];
			numVout++;

			// If the curr vertex is in and the next is out, 
			// we need to find the intersection of the edge
			// with the half-plane boundary.
			doIntersection = ((currInclusion == INSIDE) && 
						  (nextInclusion == OUTSIDE));
			}
		else
			{
			// If the current vertex is out and the next is in, 
			// then we need to find the edge/half-plane intersection
			doIntersection = (nextInclusion == INSIDE);
			}

		if (doIntersection)
			{
			// Find the intersection of the edge with the 
			// half-plane boundary and insert it into the clipped
			// polygon
			double p;
			clippedPoly.vertex[numVout][c0] = bound;
			p = (bound - poly->vertex[i][c0]) /
			       (poly->vertex[next_i][c0] - poly->vertex[i][c0]);
			clippedPoly.vertex[numVout][c1] = poly->vertex[i][c1] +
		           p * (poly->vertex[next_i][c1] - poly->vertex[i][c1]);
			numVout++;
			}


		}

	clippedPoly.numVerts = numVout;

	memcpy ((void *)poly, (void *)&clippedPoly, sizeof (Polygon));

	}  // end of clipPolyV


//
//
// Determine if a 2D polygon is entirely contained in the area of a pixel.
//
int
polyInPixel (
  Polygon *poly, 
  int pixel[2]
  )
	{
	int i;
	double lowerX, upperX, lowerY, upperY;

	lowerX = pixel[0]-0.5;
	upperX = pixel[0]+0.5;
	lowerY = pixel[1]-0.5;
	upperY = pixel[1]+0.5;

	for (i = 0; i < poly->numVerts; i++)
		{
		if ( 
			(poly->vertex[i][0] < lowerX) ||
			(poly->vertex[i][0] > upperX) ||
			(poly->vertex[i][1] < lowerY) ||
			(poly->vertex[i][1] > upperY)    )
			{
			return 0;
			}
		}

	return 1;
	}  // end of polyInPixel


//
//
// Determine if the area of a pixel is entirely contained in 
//  a 2D convex polygon.
//








int
pixelInPoly (
  int pixel[2],
  Polygon *poly
  )
	{


	// We simply determine whether each corner of the pixel area
	// is in the polygon.  Because the polygon is convex this is
	// sufficient.
	//
	// To do the point-in-polygon test, we will use the simple test 
	// of counting the number of intersections of a horizontal ray 
	// from the point with the edges of the polygon.  If the number
	// of intersections is even, then we're outside, if odd, inside.
	//
	// Calculations are somewhat simplified by the fact that the
	// edges of the polygon are axis-aligned
	//

	//
	// The intersections arrays will list the X coordinate of all
	// intersections of the two lines through the horizontal edges
	// of the pixel area.
	static double intersections[2][MAX_VERTICES];
	static int numIntersections[2];  // # of intersections for each line
	static double oldYVal[2] = {-1, -1};

	double xVal[2];  // the x coordinate of each of the lines
	double yVal[2];  // the y coordinate of each of the lines
	double d0, d1;
	unsigned int i, j, k;
	unsigned int numV_minus1;

	if (pixel == NULL)
		{
		oldYVal[0] = -1;
		oldYVal[1] = -1;
		return 0;
		}

#if 0
printf ("pixelInPoly:  pixel = %d %d\n", pixel[0], pixel[1]);
printf ("  poly = \n");
printPolygon(poly);
#endif

	xVal[0] = pixel[0] - 0.5;
	xVal[1] = pixel[0] + 0.5;
	yVal[0] = pixel[1] - 0.5;
	yVal[1] = pixel[1] + 0.5;


	//
	// We use oldYVal to check for cases where we don't have to
	// recalculate the intersection list
	//
	if (yVal[0] == oldYVal[1])
		{
		memcpy (intersections[0], intersections[1], 
					MAX_VERTICES*sizeof (double));
		numIntersections[0] = numIntersections[1];
		oldYVal[0] = yVal[0];
		numIntersections[1] = -1;
		}
	else if (yVal[1] == oldYVal[0])
		{
		memcpy (intersections[1], intersections[0], 
					MAX_VERTICES*sizeof (double));
		numIntersections[1] = numIntersections[0];
		oldYVal[1] = yVal[1];
		numIntersections[0] = -1;
		}

	numV_minus1 = poly->numVerts-1;

	// loop over each of the horizontal lines
	for (i = 0; i < 2; i++)  
	    {

	    if (yVal[i] != oldYVal[i])
		{
		oldYVal[i] = yVal[i];

	    	numIntersections[i] = 0;

	    	// loop through each of the polygon edges
	    	for (j = 0; j < poly->numVerts; j++)
		    {
		    unsigned int next_j;
		    next_j = (j+1) % poly->numVerts;

		    // Y distance from each edge end point to horizontal line
		    d0 = poly->vertex[j][1] - yVal[i];
		    d1 = poly->vertex[next_j][1] - yVal[i];

		    if (d0 == 0)
			{

			// If the line intersects the first vertex we might
			// count it as an intersection, if it intersects
			// the second vertex, we don't count it.

			// We count the vertex as an intersection only
			// if the previous vertex and the next vertex are
			// on opposite sides of horizontal line.
			// If the previous and next vertices are on the
			// same side of the horizontal line, then the 
			// the polygon boundary does not actually cross
			// the horizontal line.

			unsigned int prev_j = 
				(j+numV_minus1) % poly->numVerts;
			double dm1 = poly->vertex[prev_j][1] - yVal[i];

			if (dm1*d1 < 0)
				{
				// The polygon boundary does cross the line
				intersections[i][numIntersections[i]] =
					poly->vertex[j][0];
				numIntersections[i] ++;
				}
			}
		    else if (d1*d0 < 0)
			{
			// The two edge end points are on opposite sides
			// of the horizontal line.

			// We calculate the point of intersection and include
			// it in the list.
			intersections[i][numIntersections[i]] =
			   poly->vertex[j][0] + 
			    (poly->vertex[next_j][0] - poly->vertex[j][0]) * 
				(d0/(d0-d1));
			numIntersections[i] ++;
			}

		    } // end of loop over polygon edges
		}

	    }  // end of loop over the two horizontal lines


	// Now we count the number of intersections on the positive
	// side of each of the pixel area vertices.

	// loop over each of the horizontal lines
	for (i = 0; i < 2; i++)  
	    {
	    // count of the number of intersections on the positive side
	    // of each of the two pixel vertices on the current horizontal
	    // line.  Note that if an intersection is coincident with
	    // a pixel vertex, that vertex is automatically accepted
	    // as inside the polygon.

	    unsigned int count[2];  
	    int inside[2];  // flag for acceptance regardless of count

	    count[0] = count[1] = 0;
	    inside[0] = inside[1] = 0;

	    // Loop over all intersections on current horizontal line
	    for (j = 0; j < numIntersections[i]; j++)
		{

		// loop over each of the two pixel vertices on the current
		// horizontal line.
		for (k = 0; k < 2; k++)
			{
			double d;
			d = intersections[i][j] - xVal[k];
			if (d == 0)
				{
				// pixel vertex and intersection point 
				// are identical
				inside[k] = 1;
				}
			else if (d > 0)
				{
				// intersection point is on positive side
				// of pixel vertex
				count[k]++;
				}
			}  
		}


// printf ("pixelInPoly: line %d  count %d %d  inside %d %d\n",
// i, count[0], count[1], inside[0], inside[1]);

	    //
	    // Now we check the results of the counts and checks above.
	    // If we determine that either of the two vertices on the
	    // current horizontal line are OUTSIDE the polygon, we return
	    // with a value of false.  If both vertices are determined to
	    // be inside the polygon, we continue.
	    //
	    for (k = 0; k < 2; k++)
		{
		if ( ! inside[k] )
			{
			if ((count[k]%2) == 0)
				{
				return 0;
				}

			}
		}

	    } // end of loop over the two horizontal lines
	

	// If we've reached here, it means that all four vertices of
	// the pixel area are inside the polygon.  So we have determined
	// the the pixel area is entirely contained in the convex polygon.

	return 1;

	}  // end of pixelInPoly





//
//
// Find the area of a 2D polygon
//
double area2DPolygon (Polygon *poly)
	{
	double area = 0;
	unsigned int i, next_i, prev_i;

	for (i = 0; i < poly->numVerts; i++)
		{
		next_i = (i + 1) % poly->numVerts;
		prev_i = 
		  (i + ((unsigned int)(poly->numVerts-1))) % poly->numVerts;

		area += (poly->vertex[i][0]*poly->vertex[next_i][1] -
		         poly->vertex[i][0]*poly->vertex[prev_i][1] );
		}

	return fabs (area/2);
	}	// end of area2DPolygon

//
// 
// Find the area of a 3D triangle
//
double
areaNormal3DTriangle (Polygon *poly, double normal[3])
	{
	double edge0[3], edge1[3], crossProduct[3];
	int i;
	double len;

	if (poly->numVerts != 3)
		{
		return -1;
		}

	// Area is half the magnitude of the cross product of two
	// edge vectors.

	for (i = 0; i < 3; i++)
		{
		edge0[i] = poly->vertex[1][i] - poly->vertex[0][i];
		edge1[i] = poly->vertex[2][i] - poly->vertex[0][i];
		}

	crossProduct[0] = edge0[1]*edge1[2] - edge1[1]*edge0[2];
	crossProduct[1] = edge0[2]*edge1[0] - edge1[2]*edge0[0];
	crossProduct[2] = edge0[0]*edge1[1] - edge1[0]*edge0[1];

	len = sqrt ( 	crossProduct[0]*crossProduct[0] + 
			crossProduct[1]*crossProduct[1] + 
			crossProduct[2]*crossProduct[2] )  ;

	normal[0] = crossProduct[0]/len;
	normal[1] = crossProduct[1]/len;
	normal[2] = crossProduct[2]/len;

	return len / 2;

	}  // end of areaNormal3DTriangle





//
//
// We want to calculate the area of the overlap of the pixel and
// the 2D polygon.
//
double
pixelPolyOverlapArea (
  int pixel[2],
  Polygon *convPoly
  )
	{
	Polygon localPoly;

	memcpy (&localPoly, convPoly, sizeof(Polygon));

	if (polyInPixel (&localPoly, pixel))
		{
		// all conv verts in pixel
// printf ("polyInPixel\n");		
		return area2DPolygon (&localPoly);
		}
	else if (pixelInPoly (pixel, &localPoly))
		{
// printf ("pixelInPoly\n");
		// all pixel verts in conv
		return 1;
		}


	// We clip the polygon against each of the half-planes
	// corresponding to the edges of the pixel square

	// Note that we are going for simplicity rather than speed.

	clipPolyVH (&localPoly, 0, pixel[0]-0.5,  1);


	clipPolyVH (&localPoly, 0, pixel[0]+0.5, -1);


	clipPolyVH (&localPoly, 1, pixel[1]-0.5,  1);


	clipPolyVH (&localPoly, 1, pixel[1]+0.5, -1);


	// The resulting clipped polygon is the intersection
	// of two convex polygons, so it is itself convex.  

	// So now we compute the area of the intersection polygon.

	return area2DPolygon (&localPoly);
	}  // end of pixelPolyOverlapArea


///////////////////////////////////////////////////////////////

//
//
// Clip a 3D polygon to a half-space given by a plane equation.
//
void
clipPolygonToPlane (double planeEq[4], Polygon *poly, int vflag)
	{
	double d[MAX_VERTICES];
	Polygon clippedPoly;
	unsigned int i, next_i, numV, doIntersection;
	double p;


	if (poly->numVerts == 0)
		{
		return;
		}


	// array d gives signed distance from each seg end point to plane

	for (i = 0; i < poly->numVerts; i++)
		{

		d[i] = 	planeEq[0] * poly->vertex[i][0] +
			planeEq[1] * poly->vertex[i][1] +
			planeEq[2] * poly->vertex[i][2] +
			planeEq[3];
		}

	if (poly->numVerts == 1)
		{
		if (d[0] <= 0)
			{
			poly->numVerts = 0;
			}
		return;
		}


	numV = 0;

	// loop over vertices 
	for (i = 0; i < poly->numVerts; i++)
		{
		next_i = (i + 1) % poly->numVerts;



		if (d[i] >= 0)
			{
			// if vertex is inside or on boundary of half-space
			clippedPoly.vertexFlag[numV] = poly->vertexFlag[i];
			clippedPoly.vertex[numV][0] = poly->vertex[i][0];
			clippedPoly.vertex[numV][1] = poly->vertex[i][1];
			clippedPoly.vertex[numV][2] = poly->vertex[i][2];
			numV++;
			// need to find intersection if edge actually crosses
			doIntersection = (d[i] > 0) && (d[next_i] < 0);
			}
		else 
			{
			// need to find intersection if edge crosses
			doIntersection = (d[next_i] > 0);
			}
			
		if (doIntersection)
			{
			int j;

			// find intersection and put it inot clippedPoly
			p = -d[i]/(d[next_i] - d[i]);
			clippedPoly.vertexFlag[numV] = vflag;
			for (j = 0; j < 3; j++)
			    {
			    clippedPoly.vertex[numV][j] = poly->vertex[i][j] + 
			       p*(poly->vertex[next_i][j] - poly->vertex[i][j]);
			    }
			numV++;
			}

		}


	clippedPoly.numVerts = numV;
	memcpy (poly, &clippedPoly, sizeof (Polygon));

	}  // end of clipTriangleToPlane



//
//
// Convert a polygon in 3D world coordinates to 3D camera coordinates
//
void
worldToCameraPoly (Camera *camera, Polygon *worldPoly, Polygon *cameraPoly)
	{
	double tmpPt[3];
	double camToWorldRot[4][4];
	double worldToCamRot[4][4];
	int i;

	// Not very efficient, but who cares?

	quatToMat (camera->orientation, camToWorldRot);
	transposeMat (camToWorldRot, worldToCamRot);

	cameraPoly->numVerts = worldPoly->numVerts;

	for (i = 0; i < worldPoly->numVerts; i++)
		{
		cameraPoly->vertexFlag[i] = worldPoly->vertexFlag[i];
		tmpPt[0] = worldPoly->vertex[i][0] - camera->location[0];
		tmpPt[1] = worldPoly->vertex[i][1] - camera->location[1];
		tmpPt[2] = worldPoly->vertex[i][2] - camera->location[2];
		xformPointByMat (tmpPt, worldToCamRot, cameraPoly->vertex[i]);
		}

	}  // end of worldToCameraPoly


//
//
// Convert a polygon in 3D camera coordinates to 3D world coordinates
//
void
cameraToWorldPoly (Camera *camera, Polygon *cameraPoly, Polygon *worldPoly)
	{
	double tmpPt[3];
	double camToWorldRot[4][4];
	int i;

	worldPoly->numVerts = cameraPoly->numVerts;

	// Not very efficient, but who cares?

	quatToMat (camera->orientation, camToWorldRot);

	for (i = 0; i < worldPoly->numVerts; i++)
		{
		worldPoly->vertexFlag[i] = cameraPoly->vertexFlag[i];
		xformPointByMat (cameraPoly->vertex[i], camToWorldRot, tmpPt);
		worldPoly->vertex[i][0] = tmpPt[0] + camera->location[0];
		worldPoly->vertex[i][1] = tmpPt[1] + camera->location[1];
		worldPoly->vertex[i][2] = tmpPt[2] + camera->location[2];
		}

	}  // end of cameraToWorldPoly




//
//
// Convert a 3D polygon in camera coordinates to a 2D polygon in pixel coords
//
void
cameraToPixelPoly (
  View *view, 
  Polygon *cameraPoly, 
  Polygon *pixelPoly
  )
	{
	double botEdge, vertExtent, aspectRatio, horizExtent, leftEdge;
	int i;

	double projectedXY[2];

	botEdge = - tan (DEG_TO_RAD*view->camera.fov/2.0);
	vertExtent = 2 * tan (DEG_TO_RAD*view->camera.fov/2.0);

	aspectRatio = ((double)(view->windowSize[0]-1)) / 
				((double)(view->windowSize[1]-1));
	horizExtent = vertExtent * aspectRatio;
	leftEdge = botEdge * aspectRatio;

	pixelPoly->numVerts = cameraPoly->numVerts;
	for (i = 0; i < cameraPoly->numVerts; i++)
		{
		pixelPoly->vertexFlag[i] = cameraPoly->vertexFlag[i];

		projectedXY[0] = 
			- cameraPoly->vertex[i][0] / cameraPoly->vertex[i][2];
		projectedXY[1] = 
			- cameraPoly->vertex[i][1] / cameraPoly->vertex[i][2];

		pixelPoly->vertex[i][0] = -0.5 + ((projectedXY[0] - leftEdge) / 
					horizExtent) * (view->windowSize[0]);
		pixelPoly->vertex[i][1] = -0.5 + ((projectedXY[1] - botEdge) / 
					vertExtent) * (view->windowSize[1]);
		pixelPoly->vertex[i][2] = 0.0;
		}

	}  // end of cameraToPixelPoly


//
//
// Get the (coefficients of the) plane equations for a viewing frustum
//
void
getViewFrustumPlaneEqs (View *view, double planeEq[6][4])
	{
	double aspectRatio, horizFOV_o2;

	// We construct the plane equations for the six planes in camera coords.
	// These will be the four coefficients of the equation 
	//   Ax + By + Cz + D = 0 for each of the six planes.
	// The plane equation coefficients will be stated so that the left
	// side of the equation will be positive when an (x,y,z) point is
	// in the "inside" half-space relative to the plane.

	// Near plane
	planeEq[0][0] = 0;
	planeEq[0][1] = 0;
	planeEq[0][2] = - 1;
	planeEq[0][3] = - view->camera.near;

	// Far plane
	planeEq[1][0] = 0;
	planeEq[1][1] = 0;
	planeEq[1][2] = 1;
	planeEq[1][3] = view->camera.far;

	// All of the other planes of the frustum go through the origin
	// which means that D = 0 in the plane equation.

	// The angular field of view specifies the vertical field
	// of view for the window.  

	// Upper plane
	//
	// The unit normal to the upper plane is  (0, sin(-fov/2), cos(-fov/2))
	planeEq[2][0] = 0;
	planeEq[2][1] = - cos (DEG_TO_RAD*view->camera.fov/2.0);
	planeEq[2][2] = - sin (DEG_TO_RAD*view->camera.fov/2.0);
	planeEq[2][3] = 0;

	// Lower plane
	planeEq[3][0] = 0;
	planeEq[3][1] =   cos (DEG_TO_RAD*view->camera.fov/2.0);
	planeEq[3][2] = - sin (DEG_TO_RAD*view->camera.fov/2.0);
	planeEq[3][3] = 0;


	// So now we get the horizontal field of view (actually the
	// half angle.
	aspectRatio = ((double)(view->windowSize[0]-1)) / 
				((double)(view->windowSize[1]-1));
	horizFOV_o2 =  
		atan (tan (DEG_TO_RAD*view->camera.fov/2.0) * aspectRatio ) ;

	// Right plane
	planeEq[4][0] = - cos (horizFOV_o2);
	planeEq[4][1] = 0;
	planeEq[4][2] = - sin (horizFOV_o2);
	planeEq[4][3] = 0;

	// Left plane
	planeEq[5][0] =   cos (horizFOV_o2);
	planeEq[5][1] = 0;
	planeEq[5][2] = - sin (horizFOV_o2);
	planeEq[5][3] = 0;

	}  // get ViewFrustmPlaneEqs




int
onEdge (GLsizei windowSize[2], double a[2])
	{
	int on = 0;

	on |=  (fabs (a[0] + 0.5) < 0.00001) ? 1 : 0;
	on |=  (fabs (a[0] - (windowSize[0] - 0.5)) < 0.001) ? 2 : 0;
	on |=  (fabs (a[1] + 0.5) < 0.00001) ? 4 : 0;
	on |=  (fabs (a[1] - (windowSize[1] - 0.5)) < 0.001) ? 8 : 0;

	return on;
	}  // end of onEdge


int
onSameEdge (GLsizei windowSize[2], double a[2], double b[2])
	{
	int aOn = onEdge (windowSize, a);
	int bOn = onEdge (windowSize, b);

	return aOn & bOn;
	}  // end of onSameEdge


//
//
// Get the perimeter of the polygon in pixel and world coordinates.
// It is assumed that pixelPoly and worldPoly correspond vertex-by-vertex.
// Only the lengths of edges that are not along the edges of the window are
// counted.
//
void
getVisiblePerimeters ( 
  View *view, 
  Polygon *pixelPoly,
  Polygon *worldPoly,
  double *visiblePixelPerimeter,
  double *visibleWorldPerimeter
  )
	{
	unsigned int i, next_i;

	if (pixelPoly->numVerts != worldPoly->numVerts)
		{
		fprintf (stderr, 
			"getVisiblePerimeters: Polygon sizes don't match.\n");
		exit (-1);
		}

	*visiblePixelPerimeter = 0;
	*visibleWorldPerimeter = 0;
	for (i = 0; i < pixelPoly->numVerts; i++)
		{
		next_i = (i+1) % pixelPoly->numVerts;

		if (! onSameEdge (view->windowSize, pixelPoly->vertex[i], 
					pixelPoly->vertex[next_i]) )
		    {
		    *visiblePixelPerimeter += sqrt ( 
		      ((pixelPoly->vertex[i][0]-pixelPoly->vertex[next_i][0])*
		       (pixelPoly->vertex[i][0]-pixelPoly->vertex[next_i][0])) +
		      ((pixelPoly->vertex[i][1]-pixelPoly->vertex[next_i][1])*
		       (pixelPoly->vertex[i][1]-pixelPoly->vertex[next_i][1])) 
			);
		    *visibleWorldPerimeter += sqrt ( 
		      ((worldPoly->vertex[i][0]-worldPoly->vertex[next_i][0])*
		       (worldPoly->vertex[i][0]-worldPoly->vertex[next_i][0])) +
		      ((worldPoly->vertex[i][1]-worldPoly->vertex[next_i][1])*
		       (worldPoly->vertex[i][1]-worldPoly->vertex[next_i][1])) +
		      ((worldPoly->vertex[i][2]-worldPoly->vertex[next_i][2])*
		       (worldPoly->vertex[i][2]-worldPoly->vertex[next_i][2])) 
			);
		    }

		}
	
	}

static int
clippedToNear (View *view, Polygon *poly)
	{

	if (poly->numVerts)
	    {
	    int i;
	    double threshold;

	    threshold = -1.0001 * view->camera.near;


	    for (i = 0; i < poly->numVerts; i++)
		{
		if (poly->vertex[i][2] > threshold)
			{
			return 1;
			}

		}
	    }

	return 0;
	}

//
//
// Clip a 3D polygon to the viewing frustum.  Produce results in both
// 3D world and 2D pixel coordinates.
//
void
clipToViewingFrustum ( 
  View *view, 
  Polygon *poly, 
  Polygon *clippedWorldPoly,
  Polygon *clippedPixelPoly,
  double *visibleWorldPerimeter,
  double *visiblePixelPerimeter,
  double zExtentVisible[2],
  double normal[3]
  )
	{

	// We assume that the polygon is convex.  Thus each of the
	// clippings results in a convex polygon
	

	// The viewing frustum is defined by six planes.  We will find
	// the intersection of the polygon with each of these planes.
	//
	// We are going to transform the points into "camera" coordinates, 
	// calculate the intersections, then transform back.  This makes
	// the intersection calculations a whole lot easier.

	Polygon cameraPoly;
	double planeEq[6][4];
	int i;


	worldToCameraPoly (&(view->camera), poly, &cameraPoly);

	areaNormal3DTriangle (&cameraPoly, normal);

	getViewFrustumPlaneEqs (view, planeEq);

	for (i = 0; i < 6; i++)
		{
		clipPolygonToPlane (planeEq[i], &cameraPoly, i);
		}

	if (cameraPoly.numVerts > 0)
		{
		zExtentVisible[0] = 
		zExtentVisible[1] = cameraPoly.vertex[0][2];
		for (i = 1; i < cameraPoly.numVerts; i++)
			{
			zExtentVisible[0] = 
			  MIN (zExtentVisible[0], cameraPoly.vertex[i][2]);
			zExtentVisible[1] = 
			  MAX (zExtentVisible[1], cameraPoly.vertex[i][2]);
			}
		}
	else
		{
		zExtentVisible[0] = 
		zExtentVisible[1] = 0;
		}

	cameraToWorldPoly (&(view->camera), &cameraPoly, clippedWorldPoly);
	cameraToPixelPoly (view, &cameraPoly, clippedPixelPoly);

#if 0
printPolygon (&cameraPoly);
printPolygon (clippedWorldPoly);
printPolygon (clippedPixelPoly);
#endif

	getVisiblePerimeters (view, clippedPixelPoly, clippedWorldPoly, 
				visiblePixelPerimeter, visibleWorldPerimeter);
	}  // end of clipToViewingFrustum


/////////////////////////////////////////////////////////////

//
//
// Create a new buffer for screen pixels when the window size changes.
//
void
setupScrbuf (View *view, int *scrbufSize, unsigned char **scrbuf)
	{
	int newScrbufSize;

        newScrbufSize =
          4 * view->windowSize[0] * view->windowSize[1] * sizeof(unsigned char);
        if (*scrbufSize != newScrbufSize)
                {
                if (*scrbuf != NULL)
                        {
                        free (*scrbuf);
                        }


                *scrbufSize = newScrbufSize;
                *scrbuf = (unsigned char *) malloc (*scrbufSize);

                if (*scrbuf == NULL)
                        {
                        fprintf (stderr,
                          "setupScrbuf: "
			  "Fatal error allocating memory.\n");
                        exit (-1);
                        }
                }

	}  // end of setupScrbuf

static GLsizei CurrentWindowSize[2] = {907, 907};

void 
setupViewing (View *view)

	{
	GLdouble aspectRatio;
	double cameraMat[4][4];
	

	//
	// Doing this routine every time (rather than only when the
	// viewing changes) seems to degrade performance by only
	// about 7% on a 500x500 window (on acadie).
	//

	if ( (view->windowSize[0] > CurrentWindowSize[0]) ||
	     (view->windowSize[1] > CurrentWindowSize[1])     )
		{
		fprintf (stderr, "Specified viewport size (%d, %d) "
		  "is greater than current window size (%d, %d).\n",
		  view->windowSize[0], view->windowSize[1],
		  CurrentWindowSize[0], CurrentWindowSize[1]);
		exit (-1);
		}


   	glViewport (0, 0, view->windowSize[0], view->windowSize[1] );

	// If this divide fails because windowSize[1] == 0, then you're getting
	// what you deserve.
	aspectRatio = 
	  ((double)view->windowSize[0]-1) / ((double)view->windowSize[1]-1);
	
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();
	gluPerspective (view->camera.fov, aspectRatio, 
				view->camera.near, view->camera.far);

	glMatrixMode (GL_MODELVIEW);

	glLoadIdentity ();

	quatToMat (view->camera.orientation, cameraMat);
	gluLookAt (
			view->camera.location[0],
			view->camera.location[1],
			view->camera.location[2],
			-cameraMat[2][0],
			-cameraMat[2][1],
			-cameraMat[2][2],
			cameraMat[1][0],
			cameraMat[1][1],
			cameraMat[1][2]   );


	}  // end of setupViewing


//
//
// Render the polygon and fill the indicated buffer with the rendered pixels
//
int 
getRenderedPixels  (
  View *view,
  Polygon *poly,
  unsigned char *scrbuf)
	{
	int i;

	setupViewing ( view );

	// Begin rendering

	// Clear the frame buffer
	glClear (GL_COLOR_BUFFER_BIT);

	glColor3f (1.0, 1.0, 1.0);
	
	// glBegin (GL_POLYGON);
	// glBegin (GL_TRIANGLES);
	// glBegin (GL_TRIANGLE_STRIP);
	glBegin (GL_TRIANGLE_FAN);
	for (i = 0; i < poly->numVerts; i++)
		{
	    	GLVERTEX ( 	poly->vertex[i][0], 
				poly->vertex[i][1], 
				poly->vertex[i][2]);
		}

	glEnd ();

	glFinish ();

	// End rendering


	glReadPixels (0, 0, view->windowSize[0], view->windowSize[1], 
				GL_RGBA, GL_UNSIGNED_BYTE, scrbuf);

	return 0;
	}  // end of getRenderedPixels








//
//
// For the given view and polygon, derive the signed and unsigned errors.
//
void
getPolygonRenderErrors (
  View *view, 
  Polygon *poly, 
  double *area,
  // double *visibleArea,
  double *visiblePixelArea,
  double *visibleWorldPerimeter,
  double *visiblePixelPerimeter,
  int *numRenderedPixels,
  double *signedError, 
  double *unsignedError,
  double zExtentVisible[2],
  double normal[3]
  )
	{
	Polygon clippedWorldPoly;
	Polygon clippedPixelPoly;
	static int ScrbufSize = 0;
	static unsigned char *Scrbuf = NULL;
	int pixelCoords[2];
	double minX, minY, maxX, maxY;
	unsigned char *pixel;
	double overlap;
	double signedErrorArea;
	int i, j;
	double n[3];

	setupScrbuf (view, &ScrbufSize, &Scrbuf);

// printf ("poly before\n");
// printPolygon (poly);


	getRenderedPixels (view, poly, Scrbuf);


	clipToViewingFrustum (view, poly, &clippedWorldPoly, &clippedPixelPoly,
			visibleWorldPerimeter, visiblePixelPerimeter, 
			zExtentVisible, normal);
	


#if 0
printf ("world poly after clipping \n");
printPolygon (&clippedWorldPoly);
printf ("Pixel poly after clipping \n");
printPolygon (&clippedPixelPoly);
#endif



        *area = areaNormal3DTriangle (poly, n);
	// *visibleArea = area2DPolygon (&clippedWorldPoly);
	*visiblePixelArea = area2DPolygon (&clippedPixelPoly);


// printf ("area = %f     vpa = %f\n", area,  *visiblePixelArea);

	minX = view->windowSize[0]+1;
	minY = view->windowSize[1]+1;
	maxX = maxY = -1;

	for (i = 0; i < clippedPixelPoly.numVerts; i++)
		{
		minX = MIN (minX, clippedPixelPoly.vertex[i][0]);
		minY = MIN (minY, clippedPixelPoly.vertex[i][1]);
		maxX = MAX (maxX, clippedPixelPoly.vertex[i][0]);
		maxY = MAX (maxY, clippedPixelPoly.vertex[i][1]);
		}

	minX -=  0.5;
	minY -=  0.5;
	maxX +=  0.5;
	maxY +=  0.5;
	

	// We calculate a signedErrorArea for each pixel.
	// The error area is the area of the pixel that :
	//	a) is in the triangle but is not rendered, or
	//	b) is not in the triangle but is rendered.
	// Case (a) is considered to be a negative area, case (b) positive.
	// 

	pixelInPoly (NULL, NULL);  // resets the cached data for new poly

	*numRenderedPixels = 0;
	pixel = Scrbuf;
        for (j = 0; j < view->windowSize[1]; j++)
                {
		pixelCoords[1] = j;
                for (i = 0; i < view->windowSize[0]; i++)
                        {
			signedErrorArea = 0;
                        if (pixel[0] != 0)
                                {
				(*numRenderedPixels)++;
				// pixel (i,j) was rendered


				if ( (i >= minX) && (i <= maxX) &&
				     (j >= minY) && (j <= maxY) )
					{
					// pixel area might overlap polygon
					pixelCoords[0] = i;
					overlap = 
				  		pixelPolyOverlapArea (
							pixelCoords,
							&clippedPixelPoly);
					}
				else
					{
					// pixel area cannot overlap polygon
					overlap = 0;
					}

// printf ("overlap A = %f\n", overlap);
				// case (a), as described above
				signedErrorArea = (1 - overlap);

                                }
			else
				{
				// pixel (i, j) was not rendered
				if ( (i >= minX) && (i <= maxX) &&
				     (j >= minY) && (j <= maxY) )
					{
// printf ("AAAAA\n");
					// pixel area might overlap polygon
					pixelCoords[0] = i;
					overlap = pixelPolyOverlapArea (
							pixelCoords, 
							&clippedPixelPoly);
					}
				else
					{
// printf ("BBBBB\n");
					// pixel area cannot overlap polygon
					overlap = 0;
					}

// printf ("overlap B = %f\n", overlap);
				// case (a), as described above
				// case (b), as described above
				signedErrorArea = - overlap;
				}

  			*signedError += signedErrorArea;
  			*unsignedError += fabs (signedErrorArea);

// printf ("i = %d   j = %d   signedError %f    unsignedError %f\n",
// i, j, *signedError, *unsignedError);

                        pixel += 4;

                        }  // end of loop over i

                }  // end of loop over j


	}  // end of getPolyRenderErrors



///////////////////////////////////////////////////////////////////////

//
// In this section we are going to initialize a bunch of test
// cases for checking the trinagle rendering errors.  This will
// involve generating a set of test triangle, and test viewing
// parameters.
//

static unsigned int NumTestViews = 8;
static View *TestViews = NULL;
static unsigned int NumTestTriangles = 0;
static Polygon *TestTriangle = NULL;

static unsigned int CurrentCase = 0;


void
initRenderingTestCases ()
	{
	int i, j; 
	int iv, iwin, ifov, ior;
	int winSize[2][2] = { {128, 128}, {907, 907} };
	double fov[2] = {30.0, 150.0};
	double orientation[2][4];
	double len;

	if ((TestViews = (View *) malloc(NumTestViews * sizeof (View))) == NULL)
		{
		fprintf (stderr, "Unable to allocate memory for test views.\n");
		exit (-1);
		}

	if ((TestTriangle = (Polygon *) 
			  malloc (NumTestTriangles * sizeof (Polygon))) == NULL)
		{
		fprintf(stderr, 
			"Unable to allocate memory for test triangles.\n");
		exit (-1);
		}



	// generate views
	ustart (4339);  // seed for views

	for (ior = 0; ior < 2; ior++)
		{
		len = 0;
		for (j = 0; j < 4; j++)
			{
			orientation[ior][j] = uni ();
			len += orientation[ior][j] * orientation[ior][j];
			}

		len = sqrt (len);

		for (j = 0; j < 4; j++)
			{
			orientation[ior][j] /= len;
			}
		}


	iv = 0;
	for (iwin = 0; iwin < 2; iwin++)
	    {
	    for (ifov = 0; ifov < 2; ifov++)
		{
		for (ior = 0; ior < 2; ior++)
			{
			TestViews[iv].idNum = iv;
			TestViews[iv].windowSize[0] = winSize[iwin][0];
			TestViews[iv].windowSize[1] = winSize[iwin][1];
			TestViews[iv].camera.fov = fov[ifov];
			TestViews[iv].camera.orientation[0]=orientation[ior][0];
			TestViews[iv].camera.orientation[1]=orientation[ior][1];
			TestViews[iv].camera.orientation[2]=orientation[ior][2];
			TestViews[iv].camera.orientation[3]=orientation[ior][3];

			TestViews[iv].camera.location[0] = 0;
			TestViews[iv].camera.location[1] = 0;
			TestViews[iv].camera.location[2] = 0;
			TestViews[iv].camera.near = 0.001;
			TestViews[iv].camera.far = 20.0;

			iv++;
			}
		}
	    }
	

	// generate points
	ustart (5019);  // seed for points

	for (i = 0; i < NumTestTriangles; i++)
	    {
	    TestTriangle[i].numVerts = 3;
	    for (j = 0; j < 3; )
		{
		TestTriangle[i].vertexFlag[j] = -1;
		TestTriangle[i].vertex[j][0] = -1.0 + 2 * uni ();
		TestTriangle[i].vertex[j][1] = -1.0 + 2 * uni ();
		TestTriangle[i].vertex[j][2] = -1.0 + 2 * uni ();

		if ((TestTriangle[i].vertex[j][0]*TestTriangle[i].vertex[j][0] +
		     TestTriangle[i].vertex[j][1]*TestTriangle[i].vertex[j][1] +
		     TestTriangle[i].vertex[j][2]*TestTriangle[i].vertex[j][2] )
		     <= 1.0)
			{
			// We want only points inside the unit sphere
			j++;
			}
		}
	    }


	}  // end of initRenderingTestCases





void
ungetLastRenderingTestCase ()
	{
	if (CurrentCase > 0)
		{
		CurrentCase--;
		}
	}

int
getNextRenderingTestCase (View *view, Polygon *poly)
	{
	unsigned int currentView, currentTriangle;

	if (CurrentCase >= (NumTestViews * NumTestTriangles))
		{
		return -1;
		}

	currentView = CurrentCase / NumTestTriangles;
	currentTriangle = CurrentCase % NumTestTriangles;

	memcpy (view, TestViews+currentView, sizeof (View));
	memcpy (poly, TestTriangle+currentTriangle, sizeof (Polygon));

	CurrentCase++;


	if ((CurrentCase % 1000) == 0)
		{
		fprintf (stderr, "At case %d\n", CurrentCase);
		}

	return 0;
	}  // end of getNextRenderingTestCase




////////////////////////////////////////////////////////////////



void init (void) 
{
   glClearColor (0.0, 0.0, 0.0, 0.0);
   glShadeModel (GL_FLAT);

   initRenderingTestCases ();
}


static FILE *ViewFP = NULL;
static FILE *ErrorFP = NULL;

void display(void)
	{

	View view;
	Polygon polygon;
	int numRenderedPixels;
	double polyArea, /* visiblePolyArea ,*/ polyPixelArea, 
			visibleWorldPerimeter, visiblePixelPerimeter,
			signedError, unsignedError;
	double normal[3];
	static int lastViewId = -1;
	double zExtentVisible[2];

	// Clear out the entire window.
	glViewport (0, 0, CurrentWindowSize[0], CurrentWindowSize[1] );

	sleep (1);

	glClear (GL_COLOR_BUFFER_BIT);

	sleep (1);

	glFinish ();


	// CurrentCase = 759699;
	// CurrentCase = 773653;
	// CurrentCase = 176789;
	// CurrentCase = 57582;
	// CurrentCase = 3655;

	while (getNextRenderingTestCase (&view, &polygon) == 0)
		{

		
		zExtentVisible[0] =
		zExtentVisible[1] =
		polyArea =
		// visiblePolyArea =
		polyPixelArea =
		numRenderedPixels = 
		signedError =
		unsignedError = 0;

		getPolygonRenderErrors ( 
			&view, 
			&polygon, 
			&polyArea,
			// &visiblePolyArea,
			&polyPixelArea,
			&visibleWorldPerimeter,
			&visiblePixelPerimeter,
			&numRenderedPixels,
			&signedError,
			&unsignedError,
			zExtentVisible,
			normal );


                if (view.idNum != lastViewId)
                        {
                        printView (ViewFP, &view);
                        lastViewId = view.idNum;
                        }

		fprintf (ErrorFP,
			"%3d "
			"%lg "
			// "%lf "
			"%lg "
			"%lg %lg "
			"%d  "
			"%lg %lg "
			"%lg %lg "
			"%lg %lg %lg "
			"\n",
			view.idNum,
			polyArea,
			// visiblePolyArea,
			polyPixelArea,
			visibleWorldPerimeter,
			visiblePixelPerimeter,
			numRenderedPixels,
			signedError,
			unsignedError,
			zExtentVisible[0],
			zExtentVisible[1],
			normal[0],
			normal[1],
			normal[2]
			);

 // sleep (10);
 // exit (0);
		}

	exit (0);

	}  // end of display


void reshape (int w, int h)
{

   CurrentWindowSize[0] = w;
   CurrentWindowSize[1] = h;

}

void keyboard(unsigned char key, int x, int y)
{
   switch (key) {
      case 27:
         exit(0);
         break;
   }
}


static void
usage ()
        {
        fprintf (stderr, "\n");
        fprintf (stderr,
          "Usage: triangleRenderingErrors [glut args] numTri  viewFN  errorFN\n");
        fprintf (stderr,
         "   [glut args] See GLUT documentation.\n");
        fprintf (stderr,
         "   numTri   number of drawn drawn for each view\n");
        fprintf (stderr,
         "   viewFN   file to which a description of each view is written.\n");
        fprintf (stderr,
         "   errorFN  file to which error data is written for each "
         "triangle and each view.\n");
        fprintf (stderr, "\n");
        return;
        }



int main(int argc, char** argv)
{
   glutInit(&argc, argv);


   if (argc != 4)
        {
        fprintf (stderr, "Bad command line arguments.\n");
        usage ();
        exit (-1);
        }

   if (sscanf (argv[1], "%d", &NumTestTriangles) != 1)
        {
        fprintf (stderr, "Bad argument: < %s >.\n", argv[1]);
        usage ();
        exit (-1);
        }

   if (NumTestTriangles < 1)
        {
        fprintf (stderr, "Bad argument: < %s >.\n", argv[1]);
        usage ();
        exit (-1);
        }

   if ( (ViewFP = fopen (argv[2], "w")) == NULL)
        {
        fprintf (stderr, "Bad argument: < %s >.\n", argv[2]);
        usage ();
        exit (-1);
        }

   if ( (ErrorFP = fopen (argv[3], "w")) == NULL)
        {
        fprintf (stderr, "Bad argument: < %s >.\n", argv[3]);
        usage ();
        exit (-1);
        }






   // It seems to be faster (on acadie) to do it all in the front buffer
   glutInitDisplayMode (GLUT_SINGLE | GLUT_RGB);


   glutCreateWindow (argv[0]);

   init ();

   glutDisplayFunc(display); 
   glutReshapeFunc(reshape);
   glutKeyboardFunc(keyboard);

   glutReshapeWindow (907, 907);

   glutMainLoop();
   return 0;

}  // end of main 



